//**************************************************************************
//**
//**	##   ##    ##    ##   ##   ####     ####   ###     ###
//**	##   ##  ##  ##  ##   ##  ##  ##   ##  ##  ####   ####
//**	 ## ##  ##    ##  ## ##  ##    ## ##    ## ## ## ## ##
//**	 ## ##  ########  ## ##  ##    ## ##    ## ##  ###  ##
//**	  ###   ##    ##   ###    ##  ##   ##  ##  ##       ##
//**	   #    ##    ##    #      ####     ####   ##       ##
//**
//**	$Id: makeinfo.cpp,v 1.8 2001/11/09 14:40:32 dj_jl Exp $
//**
//**	Copyright (C) 1999-2001 Jnis Legzdi
//**
//**	This program is free software; you can redistribute it and/or
//**  modify it under the terms of the GNU General Public License
//**  as published by the Free Software Foundation; either version 2
//**  of the License, or (at your option) any later version.
//**
//**	This program is distributed in the hope that it will be useful,
//**  but WITHOUT ANY WARRANTY; without even the implied warranty of
//**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//**  GNU General Public License for more details.
//**
//**************************************************************************

#include "makeinfo.h"

void ProcessDehackedFiles(int argc, char** argv);

extern char*				sprnames[];
extern char* 				statename[];
extern char*				mt_names[];
extern char*				snd_names[];
extern char* 				flagnames1[32];
extern char* 				flagnames2[32];
extern char*				weapon_names[];
extern char*				ammo_names[];
extern state_action_info_t	StateActionInfo[];
extern int					numstates;
extern int					nummobjtypes;
extern int					numweapons;
extern int					num_sfx;
extern state_t				states[];
extern mobjinfo_t			mobjinfo[];
extern weaponinfo_t			weaponinfo[];
extern sfxinfo_t			sfx[];
extern string_def_t			strings[];
extern string_def_t			txtlumps[];
extern map_info_t			map_info[];
extern int					maxammo[];
extern int					perammo[];
extern int					numammo;
extern int					initial_health;
extern int					initial_ammo;
extern int					bfg_cells;
extern int					soulsphere_max;
extern int					soulsphere_health;
extern int					megasphere_health;
extern int					god_health;
extern int					shadow;
extern int					altshadow;

//==========================================================================
//
//	WriteHeader
//
//==========================================================================

void WriteHeader(FILE *f)
{
	fprintf(f, "//**************************************************************************\n");
	fprintf(f, "//**\n");
	fprintf(f, "//**	##   ##    ##    ##   ##   ####     ####   ###     ###\n");
	fprintf(f, "//**	##   ##  ##  ##  ##   ##  ##  ##   ##  ##  ####   ####\n");
	fprintf(f, "//**	 ## ##  ##    ##  ## ##  ##    ## ##    ## ## ## ## ##\n");
	fprintf(f, "//**	 ## ##  ########  ## ##  ##    ## ##    ## ##  ###  ##\n");
	fprintf(f, "//**	  ###   ##    ##   ###    ##  ##   ##  ##  ##       ##\n");
	fprintf(f, "//**	   #    ##    ##    #      ####     ####   ##       ##\n");
	fprintf(f, "//**\n");
	fprintf(f, "//**\t$""Id:$\n");
	fprintf(f, "//**\n");
	fprintf(f, "//**	Copyright (C) 1999-2001 Jnis Legzdi\n");
	fprintf(f, "//**\n");
	fprintf(f, "//**	This program is free software; you can redistribute it and/or\n");
	fprintf(f, "//**  modify it under the terms of the GNU General Public License\n");
	fprintf(f, "//**  as published by the Free Software Foundation; either version 2\n");
	fprintf(f, "//**  of the License, or (at your option) any later version.\n");
	fprintf(f, "//**\n");
	fprintf(f, "//**	This program is distributed in the hope that it will be useful,\n");
	fprintf(f, "//**  but WITHOUT ANY WARRANTY; without even the implied warranty of\n");
	fprintf(f, "//**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n");
	fprintf(f, "//**  GNU General Public License for more details.\n");
	fprintf(f, "//**\n");
	fprintf(f, "//**************************************************************************\n");
}

//==========================================================================
//
//	WriteFooter
//
//==========================================================================

void WriteFooter(FILE *f)
{
	fprintf(f, "//**************************************************************************\n");
	fprintf(f, "//\n");
	fprintf(f, "//\t$""Log:$\n");
	fprintf(f, "//**************************************************************************\n");
}

//==========================================================================
//
//	WriteStates
//
//==========================================================================

void WriteStates(void)
{
	FILE*		f;
	int			i;

	f = fopen("states.vc", "w");
    if (!f)
    {
    	printf("Cannot open states.vc\n");
        return;
	}

	WriteHeader(f);
	fprintf(f, "//**\n");
	fprintf(f, "//**\tStates table\n");
	fprintf(f, "//**\n");
	fprintf(f, "//**\tGenerated by makeinfo.\n");
	fprintf(f, "//**\n");
    fprintf(f, "//**************************************************************************\n");
	fprintf(f, "\n");

    for (i=1; StateActionInfo[i].fname; i++)
    {
#if 0
    	if (StateActionInfo[i].weapon_action)
        {
			fprintf(f, "void %s(player_t *player, pspdef_t *psp);\n", StateActionInfo[i].fname);
        }
		else
#endif
        {
			fprintf(f, "void %s(mobj_t *actor);\n", StateActionInfo[i].fname);
        }
    }

	fprintf(f, "\n__states__\n{\n");
//	fprintf(f, "\tS_NULL {\"\", 0, -1, NULL, S_NULL}\n");//- Dehacked fails
	for (i=0; i<numstates; i++)
    {
		fprintf(f, "\t%s {\"%s\", %d",
			statename[i], sprnames[states[i].sprite],
			states[i].frame & 0x7fff);
		if (states[i].frame & 0x8000)
        {
        	fprintf(f, " | FF_FULLBRIGHT");
        }
		if (states[i].model_name)
		{
			fprintf(f, ", \"%s\", %d", states[i].model_name, states[i].model_frame);
		}
#if 1
		if (states[i].tics == -1)
			fprintf(f, ", -1.0");
		else
			fprintf(f, ", %d.0 / 35.0", states[i].tics);
#else
			fprintf(f, ", %d", states[i].tics);
#endif
		fprintf(f, ", %s, %s", states[i].action_num ?
			StateActionInfo[states[i].action_num].fname : "NULL",
			statename[states[i].nextstate]);
		if (states[i].misc1 || states[i].misc2)
        {
			fprintf(f, ", %.1f, %.1f", (float)states[i].misc1, (float)states[i].misc2);
		}
		fprintf(f, "}\n");
    }
	fprintf(f, "}\n\n");
    WriteFooter(f);
    fclose(f);
}

//==========================================================================
//
//	WriteMobjInfo
//
//==========================================================================

void WriteMobjInfo(void)
{
	FILE*		f;
	int			i;
    int			j;
	int			firstflag;
    int			flags;
	int			flags2;
    const char *parent;

	f = fopen("mobjinfo.vc", "w");
    if (!f)
    {
    	printf("Cannot open mobjinfo.vc\n");
        return;
	}

	WriteHeader(f);
	fprintf(f, "//**\n");
	fprintf(f, "//**\tMobj info table\n");
	fprintf(f, "//**\n");
	fprintf(f, "//**\tGenerated by makeinfo.\n");
	fprintf(f, "//**\n");
    fprintf(f, "//**************************************************************************\n");
	fprintf(f, "\n");

	for (i = 0; i < nummobjtypes; i++)
    {
    	//	A standard comment before function
		fprintf(f, "//==========================================================================\n");
		fprintf(f, "//\n");
		fprintf(f, "//\t%s\n", mt_names[i]);
		fprintf(f, "//\n");
		fprintf(f, "//==========================================================================\n");
		fprintf(f, "\n");

        flags = mobjinfo[i].flags;
        flags2 = mobjinfo[i].flags2;

		//  ------------- Class declaration ------------
		if (flags & MF_MISSILE)
			parent = "Missile";
		else
			parent = "Actor";
		fprintf(f, "class %s:%s\n", mt_names[i], parent);
		fprintf(f, "{\n");

		//  ------------ Constructor -------------
		fprintf(f, "\t%s(void)\n", mt_names[i]);
		fprintf(f, "\t{\n");

		if (mobjinfo[i].classname)
	    	fprintf(f, "\t\tclassname = \"%s\";\n", mobjinfo[i].classname);

		//	Misc params
		if (mobjinfo[i].spawnhealth)
			fprintf(f, "\t\thealth = %d;\n\t\tspawnhealth = health;\n", mobjinfo[i].spawnhealth);
		if (mobjinfo[i].spawnhealth && mobjinfo[i].xdeathstate)
#ifdef NODEH
			fprintf(f, "\t\tgibshealth = -%d;\n", mobjinfo[i].spawnhealth >> 1);
#else
			fprintf(f, "\t\tgibshealth = -%d;\n", mobjinfo[i].spawnhealth);
#endif
		if (mobjinfo[i].radius)
			fprintf(f, "\t\tradius = %.1f;\n", (float)mobjinfo[i].radius / (float)FRACUNIT);
		if (mobjinfo[i].height)
			fprintf(f, "\t\theight = %.1f;\n", (float)mobjinfo[i].height / (float)FRACUNIT);
		if (mobjinfo[i].mass)
			fprintf(f, "\t\tmass = %.1f;\n", mobjinfo[i].mass == 0x7fffffff ? 99999.0 : (float)mobjinfo[i].mass);
		if (mobjinfo[i].speed)
			fprintf(f, "\t\tspeed = %.1f;\n", 35.0 * (mobjinfo[i].speed < 100 ? (float)mobjinfo[i].speed : (float)mobjinfo[i].speed / (float)FRACUNIT));
		if (mobjinfo[i].reactiontime)
        {
        	fprintf(f, "\t\tif (gameskill != sk_nightmare)\n");
			fprintf(f, "\t\t\treactiontime = %d;\n", mobjinfo[i].reactiontime);
		}
		if (mobjinfo[i].painchance)
			fprintf(f, "\t\tpainchance = %d;\n", mobjinfo[i].painchance);
		if (mobjinfo[i].damage)
			fprintf(f, "\t\tdamage = %d;\n", mobjinfo[i].damage);

        //	Flag replacements
		if (flags2 & MF2_DONTDRAW)
			fprintf(f, "\t\ttranslucency = 100;\n");
		else if (flags & MF_SHADOW)
			fprintf(f, "\t\ttranslucency = %d;\n", shadow);
		else if (flags & MF_ALTSHADOW && altshadow)
			fprintf(f, "\t\ttranslucency = %d;\n", altshadow);
#ifndef STRIFE
		else if (flags & MF_TRANSLUCENT)
			fprintf(f, "\t\ttranslucency = 33;\n");
#endif
#ifdef STRIFE
        if (flags & 0x70000000)
			fprintf(f, "\t\ttranslation = %d;\n", (flags >> 28) & 7);
#else
        if (flags & MF_TRANSLATION)
			fprintf(f, "\t\ttranslation = %d;\n", (flags & MF_TRANSLATION) >> MF_TRANSSHIFT);
#endif

		//	Clear replaced flags
        flags &= ~(MF_SHADOW|MF_NOTDMATCH);
		if (altshadow) flags &= ~MF_ALTSHADOW;
#ifdef STRIFE
		flags &= ~(0x70000000);
#else
		flags &= ~(MF_TRANSLUCENT|MF_TRANSLATION);
#endif
        flags2 &= ~(MF2_DONTDRAW);

        //	Flags
		if (flags)
        {
			fprintf(f, "\t\tflags = ");
			firstflag = 1;
    	    for (j=0; j<32; j++)
        	{
        		if (flags & (1 << j))
                {
                	if (!firstflag)
                    {
                    	fprintf(f, "|");
					}
                    firstflag = 0;
                    fprintf(f, "%s", flagnames1[j]);
                }
			}
	        fprintf(f, ";\n");
		}
		if (flags2)
        {
			fprintf(f, "\t\tflags2 = ");
			firstflag = 1;
    	    for (j=0; j<32; j++)
        	{
        		if (flags2 & (1 << j))
                {
                	if (!firstflag)
                    {
                    	fprintf(f, "|");
					}
                    firstflag = 0;
                    fprintf(f, "%s", flagnames2[j]);
                }
			}
	        fprintf(f, ";\n");
		}

		//	States
        if (mobjinfo[i].spawnstate)
			fprintf(f, "\t\tspawnstate = %s;\n", statename[mobjinfo[i].spawnstate]);
        if (mobjinfo[i].seestate)
			fprintf(f, "\t\tseestate = %s;\n", statename[mobjinfo[i].seestate]);
        if (mobjinfo[i].meleestate)
			fprintf(f, "\t\tmeleestate = %s;\n", statename[mobjinfo[i].meleestate]);
        if (mobjinfo[i].missilestate)
			fprintf(f, "\t\tmissilestate = %s;\n", statename[mobjinfo[i].missilestate]);
        if (mobjinfo[i].painstate)
			fprintf(f, "\t\tpainstate = %s;\n", statename[mobjinfo[i].painstate]);
        if (mobjinfo[i].crashstate)
			fprintf(f, "\t\tcrashstate = %s;\n", statename[mobjinfo[i].crashstate]);
        if (mobjinfo[i].deathstate)
			fprintf(f, "\t\tdeathstate = %s;\n", statename[mobjinfo[i].deathstate]);
        if (mobjinfo[i].xdeathstate)
			fprintf(f, "\t\txdeathstate = %s;\n", statename[mobjinfo[i].xdeathstate]);
        if (mobjinfo[i].raisestate)
			fprintf(f, "\t\traisestate = %s;\n", statename[mobjinfo[i].raisestate]);

		//	Sounds
        if (mobjinfo[i].seesound)
			fprintf(f, "\t\tsound_sight = GetSoundID(\"%s\");\n", sfx[mobjinfo[i].seesound].tagName);
		if (mobjinfo[i].activesound)
			fprintf(f, "\t\tsound_active = GetSoundID(\"%s\");\n", sfx[mobjinfo[i].activesound].tagName);
		if (mobjinfo[i].attacksound)
			fprintf(f, "\t\tsound_attack = GetSoundID(\"%s\");\n", sfx[mobjinfo[i].attacksound].tagName);
		if (mobjinfo[i].painsound)
			fprintf(f, "\t\tsound_pain = GetSoundID(\"%s\");\n", sfx[mobjinfo[i].painsound].tagName);
		if (mobjinfo[i].deathsound)
			fprintf(f, "\t\tsound_death = GetSoundID(\"%s\");\n", sfx[mobjinfo[i].deathsound].tagName);

		//  Effects
		if (mobjinfo[i].effects)
			fprintf(f, "\t\teffects = %s;\n", mobjinfo[i].effects);

		fprintf(f, "\t}\n");

        //  ------------ OnMapSpawn method -----------
		bool no_deathmatch = flags & MF_NOTDMATCH && mobjinfo[i].doomednum != -1;
#ifdef NODEH
		bool no_monsters = flags & MF_COUNTKILL;
#else
		bool no_monsters = flags & MF_COUNTKILL || mobjinfo[i].doomednum == 3006;
#endif
		if (no_deathmatch || no_monsters || mobjinfo[i].extra)
		{
			fprintf(f, "\n");
    	    fprintf(f, "\tvoid OnMapSpawn(mthing_t *mthing)\n");
        	fprintf(f, "\t{\n");

			if (no_deathmatch)
    	    	fprintf(f, "\t\tif (deathmatch)\n\t\t{\n\t\t\tRemoveMobjThinker(this);\n\t\t\treturn;\n\t\t}\n");
			if (no_monsters)
        		fprintf(f, "\t\tif (nomonsters)\n\t\t{\n\t\t\tRemoveMobjThinker(this);\n\t\t\treturn;\n\t\t}\n");

			//	Calling of start function
			fprintf(f, "\t\t::OnMapSpawn(mthing);\n");

			//	Static lights
			if (mobjinfo[i].extra)
				fprintf(f, "\t\t%s\n", mobjinfo[i].extra);

	        //	End of function
    	    fprintf(f, "\t}\n");
		}

		//  End of class
		fprintf(f, "};\n");
		fprintf(f, "\n");
    }

	fprintf(f, "__mobjinfo__\n{\n");
	for (i=0; i<nummobjtypes; i++)
    {
		if (mobjinfo[i].doomednum > 0)
		{
	    	fprintf(f, "\t{ %d, %s }\n", mobjinfo[i].doomednum, mt_names[i]);
		}
    }
    fprintf(f, "}\n\n");
    WriteFooter(f);
	fclose(f);
}

//==========================================================================
//
//	WriteWeaponInfo
//
//==========================================================================

void WriteWeaponInfo(void)
{
	FILE*		f;
	int			i;

	f = fopen("wpninfo.vc", "w");
    if (!f)
    {
    	printf("Cannot open wpninfo.vc\n");
        return;
	}

	WriteHeader(f);
	fprintf(f, "//**\n");
	fprintf(f, "//**\tWeaponn info table\n");
	fprintf(f, "//**\n");
	fprintf(f, "//**\tGenerated by makeinfo.\n");
	fprintf(f, "//**\n");
    fprintf(f, "//**************************************************************************\n");
	fprintf(f, "\n");

	for (i=0; i<(numweapons == 18 ? 9 : numweapons); i++)
    {
        fprintf(f, "void %s(player_t *player);\n", weapon_names[i]);
	}
    fprintf(f, "\n");

	fprintf(f, "weapon_func_t\tweaponinfo[] =\n{\n");
	for (i=0; i<(numweapons == 18 ? 9 : numweapons); i++)
    {
        fprintf(f, "\t%s,\n", weapon_names[i]);
    }
    fprintf(f, "};\n\n");

	if (!altshadow)
    {
        fprintf(f, "int weapon_ammo_type[] =\n{\n");
		for (i=0; i<(numweapons == 18 ? 9 : numweapons); i++)
    	{
        	fprintf(f, "\t%s,\n", ammo_names[weaponinfo[i].ammo]);
	    }
    	fprintf(f, "};\n\n");
	}

	for (i=0; i<(numweapons == 18 ? 9 : numweapons); i++)
    {
    	//	A standard comment before function
		fprintf(f, "//==========================================================================\n");
		fprintf(f, "//\n");
		fprintf(f, "//\t%s\n", weapon_names[i]);
		fprintf(f, "//\n");
		fprintf(f, "//==========================================================================\n");
		fprintf(f, "\n");

        //	Start of function
        fprintf(f, "void %s(player_t *player)\n", weapon_names[i]);
        fprintf(f, "{\n");

        if (altshadow)
	        fprintf(f, "\tplayer->w_mana = %s;\n", ammo_names[weaponinfo[i].ammo]);
		else
	        fprintf(f, "\tplayer->w_ammo = %s;\n", ammo_names[weaponinfo[i].ammo]);
		fprintf(f, "\tplayer->w_upstate = %s;\n", statename[weaponinfo[i].upstate]);
		fprintf(f, "\tplayer->w_downstate = %s;\n", statename[weaponinfo[i].downstate]);
		fprintf(f, "\tplayer->w_readystate = %s;\n", statename[weaponinfo[i].readystate]);
		fprintf(f, "\tplayer->w_atkstate = %s;\n", statename[weaponinfo[i].atkstate]);
		fprintf(f, "\tplayer->w_holdatkstate = %s;\n", statename[weaponinfo[i].holdatkstate]);
		fprintf(f, "\tplayer->w_flashstate = %s;\n", statename[weaponinfo[i].flashstate]);

        if (numweapons == 18)
        {
	        fprintf(f, "\tplayer->w2_ammo = %s;\n", ammo_names[weaponinfo[9+i].ammo]);
			fprintf(f, "\tplayer->w2_upstate = %s;\n", statename[weaponinfo[9+i].upstate]);
			fprintf(f, "\tplayer->w2_downstate = %s;\n", statename[weaponinfo[9+i].downstate]);
			fprintf(f, "\tplayer->w2_readystate = %s;\n", statename[weaponinfo[9+i].readystate]);
			fprintf(f, "\tplayer->w2_atkstate = %s;\n", statename[weaponinfo[9+i].atkstate]);
			fprintf(f, "\tplayer->w2_holdatkstate = %s;\n", statename[weaponinfo[9+i].holdatkstate]);
			fprintf(f, "\tplayer->w2_flashstate = %s;\n", statename[weaponinfo[9+i].flashstate]);
        }

        //	End of function
        fprintf(f, "}\n");
        fprintf(f, "\n");
	}
    WriteFooter(f);
	fclose(f);
}

#ifndef NODEH

//==========================================================================
//
//  WriteStrings
//
//==========================================================================

static void WriteStrings(void)
{
	int			i;
	FILE		*f;

	f = fopen("strings.vc", "w");
	WriteHeader(f);
	fprintf(f, "\n");
    for (i = 0; strings[i].macro; i++)
    {
		fprintf(f, "#define %s\t\t\"%s\"\n", strings[i].macro,
           	strings[i].new_val ? strings[i].new_val : strings[i].def_val);
    }
    fprintf(f, "\n");
    WriteFooter(f);
	fclose(f);
}

//==========================================================================
//
//	WriteMisc
//
//==========================================================================

static void WriteMisc(void)
{
	FILE		*f;
	int			i;

   	f = fopen("gendefs.vc", "w");

	WriteHeader(f);
    fprintf(f, "\n");
	fprintf(f, "int\t\tINITIAL_HEALTH = %d;\n", initial_health);
	fprintf(f, "int\t\tINITIAL_AMMO = %d;\n", initial_ammo);
	fprintf(f, "int\t\tBFGCELLS = %d;\n", bfg_cells);
	fprintf(f, "int\t\tSOULSPHERE_MAX = %d;\n", soulsphere_max);
	fprintf(f, "int\t\tSOULSPHERE_HEALTH = %d;\n", soulsphere_health);
	fprintf(f, "int\t\tMEGASPHERE_HEALTH = %d;\n", megasphere_health);
	fprintf(f, "int\t\tGOD_HEALTH = %d;\n", god_health);

    if (numammo)
    {
    	fprintf(f, "\nint\t\tmaxammo[] = { ");
    	for (i = 0; i < numammo; i++)
        {
			if (i) fprintf(f, ", ");
            fprintf(f, "%d", maxammo[i]);
        }
        fprintf(f, " };\n");

    	fprintf(f, "int\t\tclipammo[] = { ");
    	for (i = 0; i < numammo; i++)
        {
			if (i) fprintf(f, ", ");
            fprintf(f, "%d", perammo[i]);
        }
        fprintf(f, " };\n");
    }

    fprintf(f, "\n");
    WriteFooter(f);
	fclose(f);
}

//==========================================================================
//
//  WriteTxtLumps
//
//==========================================================================

static void WriteTxtLumps(void)
{
	int			i;
	FILE		*f;

    for (i = 0; txtlumps[i].macro; i++)
    {
		f = fopen(txtlumps[i].macro, "w");
		fprintf(f, "%s", txtlumps[i].new_val ? txtlumps[i].new_val : txtlumps[i].def_val);
		fclose(f);
    }

	f = fopen("sfxinfo.txt", "w");
    for (i = 1; i < num_sfx; i++)
    {
    	fprintf(f, "%-32s%-12s%-4d %-4d %d\n", sfx[i].tagName,
    		sfx[i].lumpname[0] ? sfx[i].lumpname : "?",
			sfx[i].priority, sfx[i].numchannels, sfx[i].changePitch);
    }
	fclose(f);

	f = fopen("sndinfo.txt", "w");
    for (i = 1; i < num_sfx; i++)
    {
    	fprintf(f, "%-32s%-12s\n", sfx[i].tagName,
    		sfx[i].lumpname[0] ? sfx[i].lumpname : "?");
    }
	fclose(f);

	if (map_info[0].name[0])
    {
    	f = fopen("mapinfo.txt", "w");
#ifdef DOOM2
    	for (i = 0; i < 32; i++)
        {
        	fprintf(f, "map MAP%02d \"%s\"\nnext MAP%02d\nSKY1 SKY%d 0\nMusic %s\n\n",
            	i + 1, map_info[i].name, (i == 29 || i == 31) ? 1 : i + 2,
            	i < 11 ? 1 : i < 20 ? 2 : 3, map_info[i].song);
        }
#else
    	for (i = 0; i < 36; i++)
        {
        	fprintf(f, "map E%dM%d \"%s\"\nnext E%dM%d\nSKY1 SKY%d 0\nMusic %s\n\n",
            	(i / 9) + 1, (i % 9) + 1, map_info[i].name,
            	(i / 9) + 1, (i % 9) > 6 ? 1 : (i % 9) + 2,
            	i < 11 ? 1 : i < 20 ? 2 : 3, map_info[i].song);
        }
#endif
        fclose(f);
    }
}

#endif

//==========================================================================
//
//	main
//
//==========================================================================

int main(int argc, char** argv)
{
#ifndef NODEH
	ProcessDehackedFiles(argc, argv);
#endif
	WriteStates();
    WriteMobjInfo();
    WriteWeaponInfo();
#ifndef NODEH
    WriteStrings();
    WriteMisc();
	WriteTxtLumps();
#endif

	return 0;
}

//**************************************************************************
//
//	$Log: makeinfo.cpp,v $
//	Revision 1.8  2001/11/09 14:40:32  dj_jl
//	Initialization in constructors
//	
//	Revision 1.7  2001/10/22 17:28:02  dj_jl
//	Removed mobjinfo index constants
//	
//	Revision 1.6  2001/10/18 17:42:19  dj_jl
//	Seperate class for missiles
//	
//	Revision 1.5  2001/10/02 17:40:48  dj_jl
//	Possibility to declare function's code inside class declaration
//	
//	Revision 1.4  2001/09/27 17:04:39  dj_jl
//	Effects and static lights in mobjinfo, mobj classes
//	
//	Revision 1.3  2001/09/20 16:33:14  dj_jl
//	Beautification
//	
//	Revision 1.2  2001/07/27 14:27:55  dj_jl
//	Update with Id-s and Log-s, some fixes
//
//**************************************************************************
